from typing import Optional

from peek_plugin_diagram._private.storage.Lookups import DispColor
from peek_plugin_diagram._private.storage.Lookups import DispLineStyle
from peek_plugin_diagram.worker.canvas_shapes.ShapeBase import Point
from peek_plugin_diagram.worker.canvas_shapes.ShapeBase import ShapeBase


class PolygonFillDirectionEnum:
    fillBottomToTop = 0
    fillBottomToTop = 1
    fillRightToLeft = 2
    fillLeftToRight = 3


class ShapePolygon(ShapeBase):
    FILL_TOP_TO_BOTTOM = 0
    FILL_BOTTOM_TO_TOP = 1
    FILL_RIGHT_TO_LEFT = 2
    FILL_LEFT_TO_RIGHT = 3

    @staticmethod
    def fillColor(disp) -> DispColor:
        return disp.get("fcl")

    @staticmethod
    def lineColor(disp) -> DispColor:
        return disp.get("lcl")

    @staticmethod
    def lineStyle(disp) -> DispLineStyle:
        return disp.get("lsl")

    @staticmethod
    def lineWidth(disp) -> DispLineStyle:
        return disp.get("w")

    @staticmethod
    def cornerRadius(disp) -> float:
        return disp.get("cr")

    @staticmethod
    def fillDirection(disp) -> int:
        fillDirection = disp.get("fd")
        if fillDirection in (
            ShapePolygon.FILL_BOTTOM_TO_TOP,
            ShapePolygon.FILL_RIGHT_TO_LEFT,
            ShapePolygon.FILL_LEFT_TO_RIGHT,
        ):
            return fillDirection

        # Else, default to Top to Bottom
        return ShapePolygon.FILL_TOP_TO_BOTTOM

    @staticmethod
    def fillPercent(disp) -> Optional[float]:
        fillPercentage = float(disp.get("fp")) if disp.get("fp") else None

        if fillPercentage is None:
            return None

        return max(0.0, min(100.0, fillPercentage))

    @staticmethod
    def isRectangle(disp) -> bool:
        return bool(disp.get("r"))

    @staticmethod
    def center(disp) -> Point:
        # TODO
        raise NotImplementedError

    @staticmethod
    def contains(disp, point: Point, margin: float) -> bool:
        raise NotImplementedError

    @staticmethod
    def geom(disp):
        return disp.get("g")



"""added CoordSet key

Peek Plugin Database Migration Script

Revision ID: bfdd4f46e293
Revises: 2c5f47322bda
Create Date: 2017-08-24 14:08:48.525689

"""

# revision identifiers, used by Alembic.
revision = "bfdd4f46e293"
down_revision = "2c5f47322bda"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "ModelCoordSet",
        sa.Column("key", sa.String(length=50), nullable=True),
        schema="pl_diagram",
    )
    op.execute(""" UPDATE "pl_diagram"."ModelCoordSet" SET "key" = "name" """)
    op.alter_column(
        "ModelCoordSet",
        "key",
        type_=sa.String(length=50),
        nullable=False,
        schema="pl_diagram",
    )


def downgrade():
    raise NotImplemented



from sqlalchemy import select
from sqlalchemy.dialects import postgresql

from peek_plugin_diagram._private.storage.Display import DispBase
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndex
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSetGridSize


gridSizeTable = ModelCoordSetGridSize.__table__
gridKeyIndexTable = GridKeyIndex.__table__
dispBaseTable = DispBase.__table__

sql = (
    select(
        [
            gridKeyIndexTable.c.gridKey,
            gridSizeTable.c.xGrid,
            gridSizeTable.c.yGrid,
        ]
    )
    .select_from(
        gridKeyIndexTable.join(
            dispBaseTable, dispBaseTable.c.id == gridKeyIndexTable.c.dispId
        ).join(
            gridSizeTable,
            gridSizeTable.c.coordSetId == gridKeyIndexTable.c.coordSetId,
        )
    )
    .where(dispBaseTable.c.key.in_([str(n) for n in range(100)]))
    .distinct()
)

print(
    str(
        sql.compile(
            dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}
        )
    )
)



from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.PluginNames import diagramObservableName
from peek_plugin_diagram._private.server.controller.BranchLiveEditController import (
    BranchLiveEditController,
)
from peek_plugin_diagram._private.server.controller.StatusController import (
    StatusController,
)
from peek_plugin_diagram._private.server.tuple_providers.BranchKeyToIdMapProvider import (
    BranchKeyToIdMapTupleProvider,
)
from peek_plugin_diagram._private.server.tuple_providers.BranchLiveEditTupleProvider import (
    BranchLiveEditTupleProvider,
)
from peek_plugin_diagram._private.server.tuple_providers.DiagramLoaderStatusTupleProvider import (
    DiagramLoaderStatusTupleProvider,
)
from peek_plugin_diagram._private.server.tuple_providers.ServerCoordSetTupleProvider import (
    ServerCoordSetTupleProvider,
)
from peek_plugin_diagram._private.server.tuple_providers.ServerLookupTupleProvider import (
    ServerLookupTupleProvider,
)
from peek_plugin_diagram._private.server.tuple_providers.ServerModelSetTupleProvider import (
    ServerModelSetTupleProvider,
)
from peek_plugin_diagram._private.storage.Lookups import DispColor
from peek_plugin_diagram._private.storage.Lookups import DispLineStyle
from peek_plugin_diagram._private.storage.Lookups import DispTextStyle
from peek_plugin_diagram._private.storage.Lookups import DispLevel
from peek_plugin_diagram._private.storage.Lookups import DispLayer
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSet, ModelSet
from peek_plugin_diagram._private.tuples.DiagramImporterStatusTuple import (
    DiagramImporterStatusTuple,
)
from peek_plugin_diagram._private.tuples.branch.BranchKeyToIdMapTuple import (
    BranchKeyToIdMapTuple,
)
from peek_plugin_diagram._private.tuples.branch.BranchLiveEditTuple import (
    BranchLiveEditTuple,
)


def makeTupleDataObservableHandler(
    ormSessionCreator,
    statusController: StatusController,
    branchLiveEditController: BranchLiveEditController,
):
    """ " Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param branchLiveEditController:
    :param ormSessionCreator: A callable that returns an SQLAlchemy session
    :param statusController: The status controller
    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=diagramObservableName, additionalFilt=diagramFilt
    )

    # Register TupleProviders here
    tupleObservable.addTupleProvider(
        DiagramImporterStatusTuple.tupleName(),
        DiagramLoaderStatusTupleProvider(statusController),
    )

    # Register TupleProviders here
    tupleObservable.addTupleProvider(
        ModelSet.tupleName(), ServerModelSetTupleProvider(ormSessionCreator)
    )

    # Register TupleProviders here
    tupleObservable.addTupleProvider(
        ModelCoordSet.tupleName(), ServerCoordSetTupleProvider(ormSessionCreator)
    )

    # Register TupleProviders here
    tupleObservable.addTupleProvider(
        BranchKeyToIdMapTuple.tupleName(),
        BranchKeyToIdMapTupleProvider(ormSessionCreator),
    )

    # Register TupleProviders here
    lookupTupleProvider = ServerLookupTupleProvider(ormSessionCreator)
    tupleObservable.addTupleProvider(DispLevel.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLayer.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispColor.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLineStyle.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispTextStyle.tupleName(), lookupTupleProvider)

    tupleObservable.addTupleProvider(
        BranchLiveEditTuple.tupleName(),
        BranchLiveEditTupleProvider(branchLiveEditController),
    )

    return tupleObservable



import logging
import math

import logging
from typing import Dict, List

from peek_plugin_diagram._private.storage.Display import (
    DispText,
    DispEllipse,
    DispGroup,
    DispGroupPointer,
)
from peek_plugin_diagram._private.storage.Lookups import DispTextStyle
from peek_plugin_diagram._private.storage.ModelSet import (
    ModelCoordSet,
    ModelCoordSetGridSize,
)

logger = logging.getLogger(__name__)


def makeGridKeysForDisp(
    coordSet: ModelCoordSet,
    disp,
    geomJson,
    textStyleById: Dict[int, DispTextStyle],
) -> List[str]:
    points = geomJson

    if geomJson is None:
        logger.critical("geomJson is None : %s ", disp)
        return []

    # If it's a text shape with no text, ignore it
    if isinstance(disp, DispText) and not disp.text:
        return []

    gridKeys = []
    for gridSize in coordSet.gridSizes:
        # CHECK Declutter
        if isinstance(disp, DispGroup) and disp.level is None:
            # Skip the check if the DispGroup has no level. It will be in all levels.
            pass

        elif 0.0 > (
            min(gridSize.max, (disp.level.maxZoom - 0.00001))
            - max(gridSize.min, disp.level.minZoom + 0.00001)
        ):
            continue

        if isinstance(disp, DispText) and _isTextTooSmall(
            disp, gridSize, textStyleById
        ):
            continue

        # If this is just a point shape/geom, then add it and continue
        if isinstance(disp, DispEllipse):
            minx, miny, maxx, maxy = _calcEllipseBounds(
                disp, points[0], points[1]
            )

        elif len(points) == 2:  # 2 = [x, y]
            # This should be a text
            if (
                not isinstance(disp, DispText)
                and not isinstance(disp, DispGroup)
                and not isinstance(disp, DispGroupPointer)
            ):
                logger.debug(
                    "TODO Determine size for disp type %s", disp.tupleType()
                )

            # Texts on the boundaries of grids could be a problem
            # They will seem them if the pan over just a little.
            gridKeys.append(
                gridSize.makeGridKey(
                    int(points[0] / gridSize.xGrid),
                    int(points[1] / gridSize.yGrid),
                )
            )
            continue

        else:
            # Else, All other shapes
            # Get the bounding box
            minx, miny, maxx, maxy = _calcBounds(points)

        # If the size is too small to see at the max zoom, then skip it
        size = math.hypot(maxx - minx, maxy - miny)
        largestSize = size * gridSize.max
        if largestSize < gridSize.smallestShapeSize:
            continue

        # Round the grid X min/max
        minGridX = int(minx / gridSize.xGrid)
        maxGridX = int(maxx / gridSize.xGrid) + 1

        # Round the grid Y min/max
        minGridY = int(miny / gridSize.yGrid)
        maxGridY = int(maxy / gridSize.yGrid) + 1

        if 50 < abs(minGridX - maxGridX):
            logger.warning(
                "Ignoring massive shape with disp.id=%s,"
                " it crosses too many horizontal grids at"
                " at gridSize.id=%s,"
                " grid count is %s",
                disp.id,
                gridSize.id,
                abs(minGridX - maxGridX),
            )
            continue

        if 50 < abs(minGridY - maxGridY):
            logger.warning(
                "Ignoring massive shape with disp.id=%s,"
                " it crosses too many vertical grids at"
                " at gridSize.id=%s,"
                " grid count is %s",
                disp.id,
                gridSize.id,
                abs(minGridY - maxGridY),
            )
            continue

        # Iterate through and create the grids.
        for gridX in range(minGridX, maxGridX):
            for gridY in range(minGridY, maxGridY):
                gridKeys.append(gridSize.makeGridKey(gridX, gridY))

    return gridKeys


def _pointToPixel(point: float) -> float:
    return point * 96 / 72


def _isTextTooSmall(
    disp,
    gridSize: ModelCoordSetGridSize,
    textStyleById: Dict[int, DispTextStyle],
) -> bool:
    """Is Text Too Small

    This method calculates the size that the text will appear on the diagram at max zoom
    for the provided gird.

    We'll only work this out based on the height

    NOTE: This must match how it's rendered PeekDispRenderDelegateText.ts
    """

    fontStyle = textStyleById[disp.textStyleId]

    if disp.textHeight:
        lineHeight = disp.textHeight
    else:
        fontSize = fontStyle.fontSize * fontStyle.scaleFactor
        lineHeight = _pointToPixel(fontSize)

    if fontStyle.scalable:
        largestSize = lineHeight * gridSize.max
    else:
        largestSize = lineHeight

    return largestSize < gridSize.smallestTextSize


def _calcEllipseBounds(disp, x, y):
    """Calculate the bounds of an ellipse"""
    # NOTE: To do this accurately we should look at the start and end angles.
    # in the interest simplicity we're not going to.
    # We'll potentially include SMALLEST_SHAPE_SIZE / 2 as well, no big deal.

    minx = x - disp.xRadius
    maxx = x + disp.xRadius

    miny = y - disp.yRadius
    maxy = y + disp.yRadius

    return minx, miny, maxx, maxy


def _calcBounds(points: List[float]):
    minx = None
    maxx = None
    miny = None
    maxy = None

    for index, val in enumerate(points):
        isY = bool(index % 2)

        if isY:
            if miny is None or val < miny:
                miny = val

            if maxy is None or maxy < val:
                maxy = val

        else:
            if minx is None or val < minx:
                minx = val

            if maxx is None or maxx < val:
                maxx = val

    return minx, miny, maxx, maxy



"""added column showForEdit

Peek Plugin Database Migration Script

Revision ID: 0db3aedfee95
Revises: 128b23798db0
Create Date: 2022-07-14 16:33:02.361090

"""

# revision identifiers, used by Alembic.
revision = "0db3aedfee95"
down_revision = "128b23798db0"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

affectedTables = (
    "DispColor",
    "DispTextStyle",
    "DispLineStyle",
    "DispLayer",
    "DispLevel",
)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    for tableName in affectedTables:
        # add blockApiUpdate
        op.add_column(
            tableName,
            sa.Column(
                "blockApiUpdate",
                sa.Boolean(),
                nullable=True,
            ),
            schema="pl_diagram",
        )
        # set 'blockApiUpdate' to false
        op.execute(
            f"""
                UPDATE pl_diagram."{tableName}"
                SET "blockApiUpdate" = false
                WHERE "blockApiUpdate" IS NULL;
        """
        )
        # set 'blockApiUpdate' NOT NULL
        op.alter_column(
            f"{tableName}",
            "blockApiUpdate",
            nullable=False,
            schema="pl_diagram",
        )

        # add showForEdit
        op.add_column(
            tableName,
            sa.Column("showForEdit", sa.Boolean(), nullable=True),
            schema="pl_diagram",
        )

        # set defaults to true to named colours below
        if tableName == "DispColor":
            op.execute(
                """
                UPDATE pl_diagram."DispColor"
                SET "showForEdit" = (
                    lower("name") in (
                        'red',
                        'green',
                        'gold',
                        'orange',
                        'cyan',
                        'white',
                        'blue'
                     )
                )
                WHERE "showForEdit" IS NULL;
            """
            )
        else:
            # set 'showForEdit' to false
            op.execute(
                f"""
                UPDATE pl_diagram."{tableName}"
                SET "showForEdit" = false
                WHERE "showForEdit" IS NULL;
        """
            )

        # set 'showForEdit' NOT NULL
        op.alter_column(
            f"{tableName}", "showForEdit", nullable=False, schema="pl_diagram"
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    for tableName in affectedTables:
        op.drop_column(f"{tableName}", "blockApiUpdate", schema="pl_diagram")
        op.drop_column(f"{tableName}", "showForEdit", schema="pl_diagram")
    # ### end Alembic commands ###



"""added wrap text at chars

Peek Plugin Database Migration Script

Revision ID: a151fd42572b
Revises: 57e88ad9d5cb
Create Date: 2023-03-16 10:46:57.388587

"""

# revision identifiers, used by Alembic.
revision = "a151fd42572b"
down_revision = "57e88ad9d5cb"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        table_name="DispTextStyle",
        column=sa.Column("wrapTextAtChars", sa.Integer(), nullable=True),
        schema="pl_diagram",
    )


def downgrade():
    op.drop_column(
        table_name="DispTextStyle",
        column_name="wrapTextAtChars",
        schema="pl_diagram",
    )



import json
import logging
from collections import defaultdict
from typing import List
from typing import Union

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_plugin_diagram._private.client.controller.CoordSetCacheController import (
    CoordSetCacheController,
)
from peek_plugin_diagram._private.client.controller.LocationIndexCacheController import (
    LocationIndexCacheController,
)
from peek_plugin_diagram._private.tuples.location_index.DispKeyLocationTuple import (
    DispKeyLocationTuple,
)
from peek_plugin_diagram._private.tuples.location_index.EncodedLocationIndexTuple import (
    EncodedLocationIndexTuple,
)
from peek_plugin_diagram._private.tuples.location_index.LocationIndexTuple import (
    LocationIndexTuple,
)
from peek_plugin_diagram._private.worker.tasks._CalcLocation import (
    dispKeyHashBucket,
)

logger = logging.getLogger(__name__)


class ClientDispKeyLocationTupleProvider(TuplesProviderABC):
    """Client Location Index Tuple Provider

    This class is used by the UIs when they have the offline location index
    switched off.

    """

    def __init__(
        self,
        locationCache: LocationIndexCacheController,
        coordSetCache: CoordSetCacheController,
    ):
        self._locationCache = locationCache
        self._coordSetCache = coordSetCache

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        modelSetKey = tupleSelector.selector["modelSetKey"]
        keys = tupleSelector.selector["keys"]

        keysByChunkKey = defaultdict(list)

        foundLocationIndexes: List[LocationIndexTuple] = []

        for key in keys:
            keysByChunkKey[dispKeyHashBucket(modelSetKey, key)].append(key)

        for chunkKey, subKeys in keysByChunkKey.items():
            chunk: EncodedLocationIndexTuple = self._locationCache.encodedChunk(
                chunkKey
            )

            if not chunk:
                logger.warning(
                    "Location index chunk %s is missing from cache", chunkKey
                )
                continue

            jsonStr = (
                Payload()
                .fromEncodedPayload(chunk.encodedLocationIndexTuple)
                .tuples[0]
                .jsonStr
            )
            locationsByKey = {i[0]: i[1:] for i in json.loads(jsonStr)}

            for subKey in subKeys:
                if subKey not in locationsByKey:
                    logger.warning(
                        "LocationIndex %s is missing from index, chunkKey %s",
                        subKey,
                        chunkKey,
                    )
                    continue

                # Reconstruct the data
                for locationJsonStr in locationsByKey[subKey]:
                    dispLocation = DispKeyLocationTuple.fromLocationJson(
                        locationJsonStr
                    )
                    foundLocationIndexes.append(dispLocation)

                    # Populate the coord set key
                    coordSet = self._coordSetCache.coordSetForId(
                        dispLocation.coordSetId
                    )

                    if coordSet is None:
                        logger.warning(
                            "Can not find coordSet with ID %s",
                            dispLocation.coordSetId,
                        )
                        continue

                    dispLocation.coordSetKey = coordSet.key

        # Create the vortex message
        return (
            Payload(filt, tuples=foundLocationIndexes)
            .makePayloadEnvelope()
            .toVortexMsg()
        )



import logging

from peek_abstract_chunked_index.private.client.controller.ACICacheControllerABC import (
    ACICacheControllerABC,
)
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.server.client_handlers.BranchIndexChunkLoadRpc import (
    BranchIndexChunkLoadRpc,
)
from peek_plugin_diagram._private.storage.branch.BranchIndexEncodedChunk import (
    BranchIndexEncodedChunk,
)
from peek_plugin_diagram._private.tuples.branch.BranchIndexUpdateDateTuple import (
    BranchIndexUpdateDateTuple,
)

logger = logging.getLogger(__name__)

clientBranchIndexUpdateFromServerFilt = dict(
    key="clientBranchIndexUpdateFromServer"
)
clientBranchIndexUpdateFromServerFilt.update(diagramFilt)


class BranchIndexCacheController(ACICacheControllerABC):
    """BranchIndex Cache Controller

    The BranchIndex cache controller stores all the chunks in memory,
    allowing fast access from the mobile and desktop devices.

    """

    _ChunkedTuple = BranchIndexEncodedChunk
    _UpdateDateTupleABC = BranchIndexUpdateDateTuple
    _chunkLoadRpcMethod = BranchIndexChunkLoadRpc.loadBranchIndexChunks
    _chunkIndexDeltaRpcMethod = BranchIndexChunkLoadRpc.loadBranchIndexDelta
    _updateFromLogicFilt = clientBranchIndexUpdateFromServerFilt
    _logger = logger



import logging

logging.getLogger("colormath.color_conversions").setLevel(logging.INFO)
logging.getLogger("colormath.color_objects").setLevel(logging.INFO)
logging.getLogger("colormath.chromatic_adaptation").setLevel(logging.INFO)

from colormath import color_diff_matrix
from colormath.color_conversions import convert_color
from colormath.color_diff import _get_lab_color1_vector
from colormath.color_diff import _get_lab_color2_matrix
from colormath.color_objects import sRGBColor
from colormath.color_objects import LabColor
from tinycss2.color3 import RGBA
from tinycss2.color3 import parse_color

logger = logging.getLogger(__name__)


def deltaECie2000(color1, color2, Kl=1, Kc=1, Kh=1):
    """
    Calculates the Delta E (CIE2000) of two colors.
    """
    color1_vector = _get_lab_color1_vector(color1)
    color2_matrix = _get_lab_color2_matrix(color2)
    delta_e = color_diff_matrix.delta_e_cie2000(
        color1_vector, color2_matrix, Kl=Kl, Kc=Kc, Kh=Kh
    )[0]
    # workaround to `numpy.asscalar()` deprecation
    return float(delta_e)


def parseCSSColor(color: str) -> RGBA:
    assert color != "None", "There is a string 'None' in the source data"
    try:
        result = parse_color(color)

    except Exception as e:
        logger.error(f"Failure to parse colour '{color}' of type {type(color)}")
        raise

    if not result:
        raise Exception(
            f"Failure to parse colour '{color} of type" f" {type(color)}"
        )

    return result


def rgbaToHexA(r: float, g: float, b: float, a: float) -> str:
    return "#{:02x}{:02x}{:02x}{:02x}".format(
        round(255 * r), round(255 * g), round(255 * b), round(255 * a)
    )


def rgbaToHex(r: float, g: float, b: float) -> str:
    return "#{:02x}{:02x}{:02x}".format(
        round(255 * r), round(255 * g), round(255 * b)
    )


def rgbToLab(r: int, g: int, b: int) -> LabColor:
    rgb = sRGBColor(rgb_r=r, rgb_g=g, rgb_b=b)
    return convert_color(rgb, LabColor, target_illuminant="d65")


def labToRgb(lighting, a, b) -> sRGBColor:
    lab = LabColor(lab_l=lighting, lab_a=a, lab_b=b)
    return convert_color(lab, sRGBColor, target_illuminant="d65")


def calculateColorDifference(color1: LabColor, color2: LabColor) -> float:
    return deltaECie2000(color1=color1, color2=color2)


def _invertColor(
    cssColor: str,
    backgroundCssColor: str,
    calibrate: bool = True,
    colorShift: float = 0.05,
) -> str:
    cssColor = parseCSSColor(cssColor)
    labColor = rgbToLab(cssColor.red, cssColor.green, cssColor.blue)

    backgroundCssColor = parseCSSColor(backgroundCssColor)
    backgroundLabColor = rgbToLab(
        backgroundCssColor.red,
        backgroundCssColor.green,
        backgroundCssColor.blue,
    )

    # ╔═══════════════╦════════════════════════════════════════╗
    # ║ Delta E Value ║               Perception               ║
    # ╠═══════════════╬════════════════════════════════════════╣
    # ║ <= 1.0        ║ Not perceptible by human eyes.         ║
    # ║ 1 - 2         ║ Perceptible through close observation. ║
    # ║ 2 - 10        ║ Perceptible at a glance.               ║
    # ║ 11 - 49       ║ Colors are more similar than opposite. ║
    # ║ 100           ║ Colors are exact opposite.             ║
    # ╚═══════════════╩════════════════════════════════════════╝
    colorDifference = calculateColorDifference(labColor, backgroundLabColor)

    if colorDifference <= 0.5 or colorDifference >= 99.999:
        # invert lighting if color is too similar to background color
        invertedLabColor = LabColor(
            lab_l=100 - labColor.lab_l,  # invert the lighting
            lab_a=labColor.lab_a,
            lab_b=labColor.lab_b,
        )
    else:
        invertedLabColor = labColor

    if calibrate and colorDifference < 99.999:
        # when inverted color looks too similar to background color
        lightingShift = min(10, colorShift * invertedLabColor.lab_l)

        newLighting = invertedLabColor.lab_l

        if backgroundLabColor.lab_l > 50:
            # tone the color with bright background
            newLighting -= lightingShift
        else:
            # tint the color with dark background
            newLighting += lightingShift

        # boundary limits
        newLighting = min(newLighting, 100)
        newLighting = max(0, newLighting)

        invertedLabColor = LabColor(
            lab_l=newLighting,
            lab_a=invertedLabColor.lab_a,
            lab_b=invertedLabColor.lab_b,
        )

    invertedRgb = labToRgb(
        invertedLabColor.lab_l, invertedLabColor.lab_a, invertedLabColor.lab_b
    )

    if cssColor.alpha == 1.0:
        return rgbaToHex(
            invertedRgb.rgb_r, invertedRgb.rgb_g, invertedRgb.rgb_b
        )

    # return invertedRgb
    return rgbaToHexA(
        invertedRgb.rgb_r,
        invertedRgb.rgb_g,
        invertedRgb.rgb_b,
        cssColor.alpha,  # take alpha from original input color
    )


def invertColor(cssColor: str, backgroundCssColor: str) -> str:
    return _invertColor(
        cssColor=cssColor,
        backgroundCssColor=backgroundCssColor,
        calibrate=True,
        colorShift=0.05,
    )



import logging
import time
from collections import defaultdict
from collections import namedtuple
from textwrap import dedent
from typing import Annotated
from typing import Dict
from typing import List
from typing import Optional

from sqlalchemy import and_
from sqlalchemy import func
from sqlalchemy import select
from sqlalchemy import text
from vortex.DeferUtil import noMainThread
from vortex.Payload import Payload

from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_diagram._private.storage.Display import DispBase
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndex
from peek_plugin_diagram._private.storage.GridKeyIndex import (
    GridKeyIndexCompiled,
)
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSet
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSetGridSize
from peek_plugin_diagram._private.storage.ModelSet import ModelSet
from peek_plugin_diagram._private.storage.branch.BranchIndex import BranchIndex
from peek_plugin_diagram._private.tuples.grid.GridTuple import GridTuple
from peek_plugin_diagram._private.worker.utils.ShapeLookupLinker import (
    ShapeLookupLinker,
)
from peek_plugin_diagram.tuples.grids.DecodedCompiledGridTuple import (
    DecodedCompiledGridTuple,
)
from peek_plugin_diagram.tuples.grids.GridKeyTuple import GridKeyTuple
from peek_plugin_diagram.worker.canvas_shapes.ShapeBase import ShapeBase

logger = logging.getLogger(__name__)


class WorkerDiagramGridApiImpl:
    _DEBUG_LOGGING = False

    @classmethod
    def getGridKeys(
        cls,
        modelSetKey: str,
        coordSetKey: str,
        boundingBox: Annotated[List[float], 4] = None,
    ) -> List[GridKeyTuple]:
        session = CeleryDbConn.getDbSession()
        try:
            query = text(
                dedent(
                    f"""
                WITH gridInfo AS (
                    SELECT
                        m1."xGrid" AS "xGrid",
                        m1."yGrid" AS "yGrid",
                        (mcs.id || '|' || m1.key || '.%')::text AS pattern
                    FROM
                        pl_diagram."ModelCoordSetGridSize" AS m1
                        JOIN pl_diagram."ModelCoordSet" AS mcs ON m1."coordSetId" = mcs.id
                        JOIN pl_diagram."ModelSet" AS ms ON ms.id = mcs."modelSetId"
                    WHERE
                        mcs."key" = '{coordSetKey}'
                        AND ms.key = '{modelSetKey}'
                    ORDER BY
                        m1."max" DESC
                    LIMIT 1)
               
                SELECT
                    "gridKey" AS gridKey,
                    (
                        SELECT
                            "xGrid"
                        FROM
                            gridInfo) AS width,
                    (
                        SELECT
                            "yGrid"
                        FROM
                            gridInfo) AS height,
                    count(*) AS shapeCount
                FROM
                    pl_diagram."GridKeyIndex" AS gi
                WHERE
                    gi."gridKey" LIKE (
                        SELECT
                            pattern
                        FROM
                            gridInfo)
                GROUP BY
                    gi."gridKey";
            """
                )
            )

            rows = session.execute(query)

            if boundingBox is not None and len(boundingBox) == 4:
                gridSizeRow = (
                    session.query(ModelCoordSetGridSize)
                    .join(
                        ModelCoordSet,
                        and_(
                            ModelCoordSet.id == ModelCoordSetGridSize.coordSetId
                        ),
                    )
                    .join(
                        ModelSet, and_(ModelSet.id == ModelCoordSet.modelSetId)
                    )
                    .filter(
                        and_(
                            ModelSet.key == modelSetKey,
                            ModelCoordSet.key == coordSetKey,
                        )
                    )
                    .order_by(ModelCoordSetGridSize.max.desc())
                    .first()
                )

                # if we have no shapes in this area of the diagram, then return
                # an array of no grid keys.
                if not gridSizeRow:
                    return []

                xGrid = gridSizeRow.xGrid
                yGrid = gridSizeRow.yGrid

                topLeftX = int(boundingBox[0] / xGrid)
                topLeftY = int(boundingBox[1] / yGrid)
                bottomRightX = int(boundingBox[2] / xGrid)
                bottomRightY = int(boundingBox[3] / yGrid)

            gridKeys = []
            for row in rows:
                _, _, xy = row[0].partition(".")  # gridKey
                x, _, y = xy.partition("x")

                x = int(x)
                y = int(y)

                if (
                    boundingBox is not None
                    and (topLeftX <= x <= bottomRightX)
                    and (topLeftY <= y <= bottomRightY)
                ):
                    g = GridKeyTuple(
                        gridKey=row[0],
                        width=float(row[1]),
                        height=float(row[2]),
                        shapeCount=int(row[3]),
                        modelSetKey=modelSetKey,
                        coordSetKey=coordSetKey,
                    )
                    gridKeys.append(g)
                elif boundingBox is None:
                    gridKeys.append(
                        GridKeyTuple(
                            gridKey=row[0],
                            width=float(row[1]),
                            height=float(row[2]),
                            shapeCount=int(row[3]),
                            modelSetKey=modelSetKey,
                            coordSetKey=coordSetKey,
                        )
                    )

            return gridKeys

        finally:
            session.close()

    @classmethod
    def getShapesByGridKeys(
        cls, gridKeys: List[str]
    ) -> Dict[str, DecodedCompiledGridTuple]:
        """get shapes filtered by grid keys in one query to database

        This guarantees states of all shapes are at the same time

        :param gridKeys: a list of gridKeys in str
        :return: a dict that maps gridKey to a list of DecodedCompiledGridTuple
        """
        s = time.monotonic()
        session = CeleryDbConn.getDbSession()
        try:
            query = session.query(GridKeyIndexCompiled).filter(
                GridKeyIndexCompiled.gridKey.in_(gridKeys)
            )

            rows = query.all()
            decodedCompiledGridTuplesByGridKey = {}

            for row in rows:
                # get encoded chunk and decode it as `Payload`
                gridTuplesPayload: Payload = row.decodedDataBlocking
                # get the GridTuple from Payload
                gridTuple: GridTuple = gridTuplesPayload.tuples[0]

                decodedCompiledGridTuplesByGridKey[
                    gridTuple.gridKey
                ] = gridTuple.toDecodedCompiledGridTuple()

            return decodedCompiledGridTuplesByGridKey
        finally:
            session.close()

    @classmethod
    def linkShapes(
        cls,
        decodedCompiledGridTuplesByGridKey: Dict[str, DecodedCompiledGridTuple],
    ):
        shapeLookupLinker = ShapeLookupLinker()
        shapeLookupLinker.loadLookups()

        for (
            gridKey,
            decodedCompiledGridTuple,
        ) in decodedCompiledGridTuplesByGridKey.items():
            shapeLookupLinker.linkShapesToLookups(
                decodedCompiledGridTuple.shapes
            )

    @classmethod
    def compileShapes(
        cls,
        decodedCompiledGridTuples,
        levelsOrderedByOrder,
        layersOrderedByOrder,
        shapeSelectionZoomLevel,
        enabledLayerKeys: Optional[list[str]] = None,
        enabledBranchKeys: Optional[list[str]] = None,
    ):
        """an equivalent of PeekCanvasModel._compileDisps()

        :param decodedCompiledGridTuples: a list of decodedCompiledGridTuples
                which contains compiled shapes in a grid
        :param levelsOrderedByOrder: a list of levels already ordered by
        tuplefield `order`
        :param layersOrderedByOrder: a list of layers already ordered by
        tuplefield `order`
        :param shapeSelectionZoomLevel: a zoom level to filter out shapes
        that are displayed above the zoom level
        :return: a list of dicts of shapes that are ready for factory renderer
        """
        # levelsOrderedByOrder = cls._getLevelsOrderedByOrder(
        #     coordSetKey=coordSetKey
        # )
        #
        # layersOrderedByOrder = cls._getLayersOrderedByOrder(
        #     modelSetKey=modelSetKey
        # )

        dispHashIdsAdded = set()
        viewingBranchesActive = bool(enabledBranchKeys)

        # These are overrides
        enabledLayerKeys = set(enabledLayerKeys) if enabledLayerKeys else None

        if not enabledBranchKeys:
            branchIdsActive = set()
        else:
            session = CeleryDbConn.getDbSession()
            try:
                branchIdsActive = set(
                    [
                        o.id
                        for o in session.query(BranchIndex.id).filter(
                            BranchIndex.key.in_(enabledBranchKeys)
                        )
                    ]
                )
            finally:
                session.close()

        for gridOrBranch in decodedCompiledGridTuples:
            for shape in gridOrBranch.shapes:
                # TODO Restrict for Branch Stage
                # If the branch is showing, and it replaces a hash,
                # then add the hash it replaces
                replacesHashId = ShapeBase.replacesHashId(shape)
                if (
                    ShapeBase.branchId(shape) in branchIdsActive
                    and replacesHashId
                ):
                    dispHashIdsAdded.add(replacesHashId)

        dispIndexByGridKey = defaultdict(int)
        disps = []
        for level in levelsOrderedByOrder:
            for layer in layersOrderedByOrder:
                # If override layer enableds have been defined, then use them
                if enabledLayerKeys:
                    if layer.key not in enabledLayerKeys:
                        continue

                elif not layer.visible:
                    # Else Use enabled from the layers provided
                    continue

                for decodedCompiledGridTuple in decodedCompiledGridTuples:
                    gridKey = decodedCompiledGridTuple.gridKey

                    # default is 0 if key doesn't exist
                    nextIndex = dispIndexByGridKey[gridKey]

                    shapes = decodedCompiledGridTuple.shapes

                    if nextIndex >= len(shapes):
                        continue

                    # while nextIndex < len(shapes):
                    for i in range(nextIndex, len(shapes), 1):
                        nextIndex = i
                        shape = shapes[i]

                        # Filter out overlay disps if we need to
                        if viewingBranchesActive and ShapeBase.isOverlay(shape):
                            continue

                        # for idx, shape in enumerate(shapes):
                        # Level first, as per the sortDisps function
                        shapeLevel = ShapeBase.level(shape)

                        if shapeLevel.order < level.order:
                            continue

                        if level.order < shapeLevel.order:
                            break

                        # Then Layer
                        shapeLayer = ShapeBase.layer(shape)
                        if shapeLayer.order < layer.order:
                            continue

                        if layer.order < shapeLayer.order:
                            break

                        # If we should filter based on zoom level, then do that.
                        if not (
                            shapeSelectionZoomLevel is None
                            or shapeLevel.minZoom
                            <= shapeSelectionZoomLevel
                            <= shapeLevel.maxZoom
                        ):
                            continue

                        # If the disp has already been added or is being replaced
                        # by a branch, then skip this one
                        if ShapeBase.hashId(shape) in dispHashIdsAdded:
                            continue

                        # Is the branch showed from the "View Branches" menu
                        isBranchViewable = (
                            ShapeBase.branchId(shape) in branchIdsActive
                        )

                        # If this is not apart of a branch, or ...
                        if (
                            ShapeBase.branchId(shape) is None
                            or isBranchViewable
                        ):
                            disps.append(shape)
                            dispHashIdsAdded.add(ShapeBase.hashId(shape))

                    dispIndexByGridKey[gridKey] = nextIndex

        return disps

    @classmethod
    def getGridKeyTuplesFromShapeKeys(
        cls, modelSetKey, coordSetKey, shapeKeys, smallestGridKeySize=False
    ) -> List[GridKeyTuple]:
        _GridData = namedtuple("_GridData", ("gridKey", "width", "height"))

        noMainThread()

        gridKeyIndexTable = GridKeyIndex.__table__
        dispBaseTable = DispBase.__table__

        # Ensure there are no duplicate keys or empty keys
        shapeKeys = list(set(filter(lambda k: k, shapeKeys)))

        ormSession = CeleryDbConn.getDbSession()
        try:
            coordSetId = (
                ormSession.query(ModelCoordSet.id)
                .filter(
                    ModelCoordSet.key == coordSetKey
                    and ModelCoordSet.modelSet.key == modelSetKey
                )
                .scalar()
            )

            gridKeysIterable = set()

            # Start a transaction so we have consistent results.
            CHUNK_SIZE = 1000

            for i in range(0, len(shapeKeys), CHUNK_SIZE):
                chunkedShapeKeys = shapeKeys[i : i + CHUNK_SIZE]
                sql = (
                    select([gridKeyIndexTable.c.gridKey])
                    .select_from(
                        gridKeyIndexTable.join(
                            dispBaseTable,
                            dispBaseTable.c.id == gridKeyIndexTable.c.dispId,
                        )
                    )
                    .where(dispBaseTable.c.key.in_(chunkedShapeKeys))
                    .distinct()
                )

                gridKeysIterable.update(
                    [row.gridKey for row in ormSession.execute(sql)]
                )

            # Query for the grid size data
            gridKeySizesByKey = {
                gs.key: gs
                for gs in ormSession.query(ModelCoordSetGridSize).filter(
                    ModelCoordSetGridSize.coordSetId == coordSetId
                )
            }

            # If we just want the smallest grids, filter for them
            if smallestGridKeySize:
                maxKey = max([i.key for i in gridKeySizesByKey.values()])

                ## Filter the grid sizes
                gridKeySizesByKey = {maxKey: gridKeySizesByKey[maxKey]}

                # Filter the grids
                startsWithStr = ModelCoordSetGridSize.makeGridKeyStartsWith(
                    coordSetId, maxKey
                )

                gridKeysIterable = list(
                    filter(
                        lambda gridKey: gridKey.startswith(startsWithStr),
                        gridKeysIterable,
                    )
                )

            # Query for the shape counts and create the GridKeyTuple
            sql = (
                select(
                    [
                        gridKeyIndexTable.c.gridKey,
                        func.count(gridKeyIndexTable.c.gridKey).label(
                            "shapeCount"
                        ),
                    ]
                )
                .where(gridKeyIndexTable.c.gridKey.in_(gridKeysIterable))
                .group_by(gridKeyIndexTable.c.gridKey)
            )

            gridKeyTuples = []

            for row in ormSession.execute(sql):
                gridSize = gridKeySizesByKey[
                    ModelCoordSetGridSize.gridSizeKeyFromGridKey(row.gridKey)
                ]

                gridKeyTuples.append(
                    GridKeyTuple(
                        gridKey=row.gridKey,
                        width=float(gridSize.xGrid),
                        height=float(gridSize.yGrid),
                        shapeCount=row.shapeCount,
                        modelSetKey=modelSetKey,
                        coordSetKey=coordSetKey,
                    )
                )

            return gridKeyTuples

        finally:
            ormSession.close()

    @classmethod
    def filterShapesByShapeKey(
        cls, shapes: list[dict], filterForKeys: set[str]
    ) -> list[dict]:
        assert isinstance(filterForKeys, set), (
            "Please convert filterForKeys "
            "to a set before passing it to filterShapesByShapeKey"
        )
        if cls._DEBUG_LOGGING:
            logger.debug(
                "filterShapesByShapeKey: Started,"
                " with %s shapes and %s filter keys",
                len(shapes),
                len(filterForKeys),
            )

        # Index the shapes so we know which shapes belong to which group id
        shapesByGroupId: dict[int, list[tuple[int, dict]]] = defaultdict(list)
        for index, shape in enumerate(shapes):
            groupId = ShapeBase.groupId(shape)
            if groupId is not None:
                shapesByGroupId[groupId].append((index, shape))

        if cls._DEBUG_LOGGING:
            logger.debug(
                "filterShapesByShapeKey: Completed indexing shapes by groupId, We've found %s groupIds",
                len(shapesByGroupId),
            )

        # This is all the shapes we want to collect
        shapeIndexes = set()

        # This is the list of shape groups we need to include the group
        # members for
        groupShapeParentIds = []

        # First, iterate over the shapes and filter for the shapes by keys
        for index, shape in enumerate(shapes):
            if ShapeBase.key(shape) in filterForKeys:
                shapeIndexes.add(index)
                groupShapeParentIds.append(ShapeBase.id(shape))

                groupId = ShapeBase.groupId(shape)
                if groupId:
                    groupShapeParentIds.append(groupId)

        if cls._DEBUG_LOGGING:
            logger.debug(
                "filterShapesByShapeKey: Completed filtering by key, %s shapes selected",
                len(shapeIndexes),
            )

        # Next we include the shapes that are group memebers of the shapes in
        # the last loop.
        # We also include those shapes in our next check and include those
        # shapes members - groups can have groups in them.
        loop = 1
        while groupShapeParentIds:
            nextGroupShapeParentIds = []

            for groupId in groupShapeParentIds:
                memberIndexShapeTuples = shapesByGroupId.get(groupId, [])
                for index, shape in memberIndexShapeTuples:
                    shapeIndexes.add(index)
                    nextGroupShapeParentIds.append(ShapeBase.id(shape))

            if cls._DEBUG_LOGGING:
                logger.debug(
                    "filterShapesByShapeKey: Completed filtering for"
                    " level %s group members"
                    " of %s groups,"
                    " found %s shapes",
                    loop,
                    len(groupShapeParentIds),
                    len(nextGroupShapeParentIds),
                )

            groupShapeParentIds = nextGroupShapeParentIds
            loop += 1

        if cls._DEBUG_LOGGING:
            logger.debug(
                "filterShapesByShapeKey: Completed, returning %s shapes.",
                len(shapeIndexes),
            )

        # return sortedFilteredShapes
        return [shapes[index] for index in sorted(shapeIndexes)]



from unittest import TestCase

from peek_plugin_diagram.tuples.ColorUtil import _invertColor


class TestColorInversion(TestCase):
    def setUp(self):
        pass

    def test_twoWayInversion(self):
        black = "#000000"
        white = "#ffffff"

        to = _invertColor(white, black, calibrate=False)
        back = _invertColor(to, white, calibrate=False)
        self.assertEqual(back, white)

    def test_cssColorHexShort(self):
        shortBlack = "#000"
        black = "#000000"
        white = "#ffffff"

        to = _invertColor(shortBlack, white, calibrate=False)
        back = _invertColor(to, black, calibrate=False)
        self.assertEqual(back, black)

    def test_cssColorLiteral(self):
        whiteLiteral = "white"
        black = "#000000"
        white = "#ffffff"

        to = _invertColor(whiteLiteral, black, calibrate=False)
        back = _invertColor(to, white, calibrate=False)
        self.assertEqual(back, white)

    def test_alpha(self):
        blackWithAlpha = "#000000ee"
        black = "#000000ff"
        white = "#ffffffff"

        to = _invertColor(blackWithAlpha, black, calibrate=False)
        back = _invertColor(to, white, calibrate=False)
        self.assertEqual(back, blackWithAlpha)

    def test_calibrate(self):
        black = "#000011"
        white = "#ffffff"

        to = _invertColor(black, white, calibrate=True, colorShift=0.05)
        self.assertEqual(to, "#020011")

    def test_colorIsTotalOppositeColor(self):
        # TODO: update ALL tests
        black = "#000000"
        whiteBackground = "#ffffff"

        to = _invertColor(
            black, whiteBackground, calibrate=True, colorShift=0.05
        )
        self.assertEqual(to, whiteBackground)



import logging
from typing import Union

from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.Payload import Payload
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_plugin_diagram._private.client.controller.GridCacheController import (
    GridCacheController,
)

logger = logging.getLogger(__name__)


class GridCacheIndexTupleProvider(TuplesProviderABC):
    def __init__(self, gridCacheController: GridCacheController):
        self._gridCacheController = gridCacheController

    @inlineCallbacks
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        emptyPayloadVortexMsg = PayloadEnvelope(
            filt, encodedPayload=Payload().toEncodedPayload()
        ).toVortexMsg()

        index = tupleSelector.selector.get("index")

        if index is None:
            # Make it backwards compatible with v3.3.0 and below
            start = tupleSelector.selector.get("start")
            count = tupleSelector.selector.get("count")

            if count is not None and start is not None:
                count = count - start

            if count is None or count != 5000:
                return emptyPayloadVortexMsg

            index = start / count

        try:
            encodedPayload = (
                self._gridCacheController.offlineUpdateDateByChunkKeyPayload(
                    index
                )
            )
        except IndexError:
            return emptyPayloadVortexMsg

        payloadEnvelope = PayloadEnvelope(filt, encodedPayload=encodedPayload)
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        return vortexMsg
